import asyncio
import datetime
import functools
import logging
import time
from typing import Text

import requests
import requests.adapters
from requests.packages.urllib3.util import retry

from gamla import functional


def _time_to_readable(time_s: float):
    return datetime.datetime.fromtimestamp(time_s)


def _request_id(name, args, kwargs) -> Text:
    args_str = str(args)[:50]
    kwargs_str = str(kwargs)[:50]
    return f"{name}, args: {args_str}, kwargs: {kwargs_str}"


def _async_timeit(f):
    @functools.wraps(f)
    async def wrapper(*args, **kwargs):
        req_id = _request_id(f.__name__, args, kwargs)
        start = time.time()
        logging.info(f"{req_id} started at {_time_to_readable(start)}")
        result = await f(*args, **kwargs)
        finish = time.time()
        elapsed = finish - start
        logging.info(
            f"{req_id} finished at {_time_to_readable(finish)}, took {elapsed}"
        )
        return result

    return wrapper


def timeit(f):
    if asyncio.iscoroutinefunction(f):
        return _async_timeit(f)

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        req_id = _request_id(f.__name__, args, kwargs)
        start = time.time()
        logging.info(f"{req_id} started at {_time_to_readable(start)}")
        result = f(*args, **kwargs)
        finish = time.time()
        elapsed = finish - start
        logging.info(
            f"{req_id} finished at {_time_to_readable(finish)}, took {elapsed}"
        )
        return result

    return wrapper


def requests_with_retry(retries: int = 3) -> requests.Session:
    session = requests.Session()
    adapter = requests.adapters.HTTPAdapter(
        max_retries=retry.Retry(
            total=retries, backoff_factor=0.1, status_forcelist=(500, 502, 504)
        )
    )
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


def batch_calls(f, timeout=20):
    """Batches single call into one request.

    Turns `f`, a function that gets a `tuple` of independent requests, into a function
    that gets a single request.
    """
    queue = {}

    async def make_call():
        await asyncio.sleep(0.1)
        if not queue:
            return
        promises = tuple(queue.values())
        requests = tuple(queue.keys())
        queue.clear()
        try:
            for promise, result in zip(promises, (await f(requests))):
                if promise.done() or promise.cancelled():
                    continue
                promise.set_result(result)
        except Exception as exception:
            for promise in promises:
                if promise.done() or promise.cancelled():
                    continue
                promise.set_exception(exception)

    async def wrapped(hashable_input):
        if hashable_input in queue:
            return await asyncio.wait_for(queue[hashable_input], timeout=timeout)
        async_result = asyncio.Future()
        # Check again because of context switch due to the creation of `asyncio.Future`.
        # TODO(uri): Make sure this is needed.
        if hashable_input in queue:
            return await asyncio.wait_for(queue[hashable_input], timeout=timeout)
        queue[hashable_input] = async_result
        asyncio.create_task(make_call())
        return await asyncio.wait_for(async_result, timeout=timeout)

    return wrapped


def queue_identical_calls(f):
    # Note that pending grows infinitely large, this assumes we cache the results
    # anyway after this decorator, so we at most double the memory consumption.
    pending = {}

    @functools.wraps(f)
    async def wrapped(*args, **kwargs):
        key = functional.make_call_key(args, kwargs)
        if key not in pending:
            pending[key] = asyncio.Future()
            pending[key].set_result(await f(*args, **kwargs))
        return await asyncio.wait_for(pending[key], timeout=20)

    return wrapped



import asyncio
import functools
import hashlib
import inspect
import itertools
import json
import logging
from typing import Callable, Dict, Iterable, Text, Type

import gevent
import toolz
from gevent import pool
from toolz import curried


@toolz.curry
def dict_get(dict_obj: Dict, default, key):
    return dict_obj.get(key, default)


def do_if(condition, fun):
    return curried.do(curried_ternary(condition, fun, toolz.identity))


def check(condition, exception):
    return do_if(toolz.complement(condition), make_raise(exception))


def bifurcate(*funcs):
    """Serially runs each function on tee'd copies of `input_generator`."""

    def inner(input_generator):
        return toolz.pipe(
            zip(funcs, itertools.tee(input_generator, len(funcs))),
            curried.map(star(lambda f, generator: f(generator))),
            tuple,
        )

    return inner


def singleize(func: Callable) -> Callable:
    def wrapped(some_input):
        if isinstance(some_input, tuple):
            return func(some_input)
        return toolz.first(func((some_input,)))

    async def wrapped_async(some_input):
        if isinstance(some_input, tuple):
            return await func(some_input)
        return toolz.first(await func((some_input,)))

    if inspect.iscoroutinefunction(func):
        return wrapped_async
    return wrapped


def wrapped_partial(func: Callable, *args, **kwargs) -> Callable:
    partial_func = functools.partial(func, *args, **kwargs)
    functools.update_wrapper(partial_func, func)
    return partial_func


@toolz.curry
def anymap(f: Callable, it: Iterable):
    return any(map(f, it))


def anyjuxt(*funcs):
    return toolz.compose(any, toolz.juxt(*funcs))


@toolz.curry
def allmap(f: Callable, it: Iterable):
    return all(map(f, it))


def alljuxt(*funcs):
    return toolz.compose(all, toolz.juxt(*funcs))


def ignore_input(inner):
    def ignore_and_run(*args, **kwargs):
        return inner()

    return ignore_and_run


def curried_ternary(condition, f_true, f_false):
    def inner(*args, **kwargs):
        return (
            f_true(*args, **kwargs)
            if condition(*args, **kwargs)
            else f_false(*args, **kwargs)
        )

    return inner


def make_raise(exception):
    def inner():
        raise exception

    return ignore_input(inner)


def translate_exception(func, exc1, exc2):
    """`func` is assumed to be unary."""
    return toolz.excepts(exc1, func, make_raise(exc2))


@functools.lru_cache(maxsize=None)
def compute_stable_json_hash(item) -> Text:
    return hashlib.sha1(
        json.dumps(
            json.loads(item.to_json()), sort_keys=True, separators=(",", ":")
        ).encode("utf-8")
    ).hexdigest()


def star(function: Callable) -> Callable:
    return lambda x: function(*x)


@toolz.curry
def _assert_f_output_on_inp(f, inp):
    assert f(inp)


def assert_that(f):
    return curried.do(_assert_f_output_on_inp(f))


_GLOBAL_POOL = pool.Group()


async def apipe(val, *funcs):
    for f in funcs:
        val = f(val)
        if inspect.isawaitable(val):
            val = await val
    return val


def acompose(*funcs):
    async def composed(*args, **kwargs):
        if args:
            inp = toolz.first(args)
        else:
            inp = toolz.first(kwargs.values())
        for f in reversed(funcs):
            inp = f(inp)
            if inspect.isawaitable(inp):
                inp = await inp
        return inp

    return composed


def acompose_left(*funcs):
    return acompose(*reversed(funcs))


async def materialize(async_generator):
    elements = []
    async for element in async_generator:
        elements.append(element)
    return tuple(elements)


@toolz.curry
async def amap(f, it):
    return await asyncio.gather(*map(f, it))


@toolz.curry
async def aexcepts(exception_type, func, handler, x):
    try:
        return await func(x)
    except exception_type as error:
        return handler(error)


@toolz.curry
async def mapa(f, it):
    async for element in it:
        yield f(element)


async def aconcat(async_generators):
    async for g in async_generators:
        for x in g:
            yield x


def ajuxt(*funcs):
    async def ajuxt_inner(x):
        results = []
        for f in funcs:
            result = f(x)
            if inspect.isawaitable(result):
                result = await result
            results.append(result)
        return tuple(results)

    return ajuxt_inner


@toolz.curry
async def afilter(func, it):
    it = tuple(it)
    results = await amap(func, it)
    return toolz.pipe(
        zip(it, results), curried.filter(toolz.second), curried.map(toolz.first)
    )


def afirst(*funcs, exception_type):
    async def afirst_inner(x):
        for f in funcs:
            try:
                result = f(x)
                if inspect.isawaitable(result):
                    result = await result
                return result
            except exception_type:
                pass
        raise exception_type

    return afirst_inner


@toolz.curry
def pmap(f, it):
    # The `tuple` is for callers convenience (even without it, the pool is eager).
    return tuple(_GLOBAL_POOL.map(f, it))


@toolz.curry
def pfilter(f, it):
    return toolz.pipe(
        it,
        bifurcate(pmap(f), curried.map(toolz.identity)),
        zip,
        curried.filter(toolz.first),
        curried.map(toolz.second),
    )


def pfirst(*funcs, exception_type):
    """Parallel+lazy iterator.

    This is used for when we want to start everything in parallel, but return a value
    on the first successful result.
    """
    failure_value = "gamla-value-signifying-failure"

    def inner(*args, **kwargs):
        return toolz.pipe(
            funcs,
            # Prepare runs for each function on the given input, wrapping them for
            # failure as gevent treats an exception within a greenlet as a real error.
            curried.map(
                lambda f: gevent.spawn(
                    toolz.excepts(exception_type, f, lambda _: failure_value),
                    *args,
                    **kwargs
                )
            ),
            # Materialize to actually start the requests in parallel.
            tuple,
            # Don't wait for all to finish, start examining the requests by order.
            curried.map(lambda promise: promise.get()),
            curried.filter(lambda result: result != failure_value),
            translate_exception(next, StopIteration, exception_type),
        )

    return inner


def first(*funcs, exception_type: Type[Exception]):
    def inner(*args, **kwargs):
        for func in funcs:
            try:
                return func(*args, **kwargs)
            except exception_type:
                pass
        raise exception_type

    return inner


logger = curried.do(logging.info)


def log_text(text: Text):
    return curried.do(lambda _: logging.info(text))


# To get a unique caching key for each function invocation, we take `args` and `items()`
# of `kwargs` and sort them (by keys), while also marking the beginning of `kwargs`.
# Inspired by: http://code.activestate.com/recipes/578078/ (python LRU cache
# implementation).
def make_call_key(args, kwargs):
    """Stable id for function calls, can be used for caching."""
    key = args
    if kwargs:
        key += "##kwargs##", tuple(sorted(kwargs.items()))
    return key



from gamla import graph


def test_get_connectivity_components():
    assert list(
        graph.get_connectivity_components({1: [2], 2: [1, 3], 3: [2], 4: [5], 5: [4]})
    ) == [frozenset({1, 2, 3}), frozenset({4, 5})]


def test_cliques_to_graph():
    assert graph.cliques_to_graph([{1, 2}, {2, 3}, {7, 8}]) == {
        1: frozenset({2}),
        2: frozenset({1, 3}),
        3: frozenset({2}),
        7: frozenset({8}),
        8: frozenset({7}),
    }



import itertools
from typing import Any, Callable, Dict, FrozenSet, Iterable, Text, Tuple

import toolz
from toolz import curried
from toolz.curried import operator

from gamla import functional


@toolz.curry
def graph_traverse(source: Any, get_neighbors: Callable) -> Iterable:
    """BFS over a graph, yielding unique nodes."""
    seen = set()
    queue = [source]
    while queue:
        current = queue.pop()
        yield current
        seen.add(current)
        for node in get_neighbors(current):
            if node not in seen:
                queue = [node] + queue


def traverse_graph_by_radius(
    source: Any, get_neighbors: Callable, radius: int
) -> Iterable:
    """Like `graph_traverse`, but does not traverse farther from given `radius`."""

    def get_neighbors_limiting_radius(
        current_and_distance: Tuple[Text, int]
    ) -> Iterable[Tuple[Text, int]]:
        current, distance = current_and_distance
        if distance < radius:
            yield from map(
                lambda neighbor: (neighbor, distance + 1), get_neighbors(current)
            )

    return map(
        toolz.first,
        graph_traverse(source=(source, 0), get_neighbors=get_neighbors_limiting_radius),
    )


edges_to_graph = toolz.compose(
    curried.valmap(toolz.compose(frozenset, curried.map(toolz.second))),
    curried.groupby(toolz.first),
)


def cliques_to_graph(cliques: Iterable[Iterable]) -> Dict[Any, FrozenSet[Any]]:
    return toolz.pipe(
        cliques,
        curried.mapcat(lambda clique: itertools.permutations(clique, r=2)),
        edges_to_graph,
    )


def get_connectivity_components(graph: Dict) -> Iterable[FrozenSet]:
    """Graph is assumed to undirected, so each edge must appear both ways."""
    nodes_left = frozenset(graph)
    while nodes_left:
        result = frozenset(
            graph_traverse(
                source=toolz.first(nodes_left),
                get_neighbors=toolz.compose(
                    curried.filter(operator.contains(nodes_left)), graph.get
                ),
            )
        )
        yield result
        nodes_left -= result


@toolz.curry
def groupby_many(f, it):
    return toolz.pipe(
        it,
        curried.mapcat(
            toolz.compose_left(
                lambda element: (f(element), [element]),
                functional.star(itertools.product),
            )
        ),
        edges_to_graph,
    )



import setuptools

with open("README.md", "r") as fh:
    _LONG_DESCRIPTION = fh.read()


setuptools.setup(
    name="gamla",
    version="0.0.23",
    long_description=_LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=setuptools.find_namespace_packages(),
    install_requires=["toolz", "gevent", "requests", "pytest", "pytest-asyncio"],
)
